/**
 *  @file    mc.c
 *  @author  Guillaume Bour. 2002
 
 *  @version 0.1
 *    @date  01/02/2002

 *  middle-code for the logo compiler.
 */
/*      Copyright (C) 2002 Guillaume Bour
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *  
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *  
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 59 Temple Place - Suite 330, Boston, 
 *      MA 02111-1307, USA.
 */
#include <string.h>
#include <stdarg.h>
#include <gtk/gtk.h>
#include "variable.h"
#include "literal.h"
#include "instr.h"
#include "proc.h"
#include "middle-code.h"

#include "turtle.h"
#include "type.h"
#include "debug.h"

/*>>> here is the declaration of the turtle            */
struct s_turtle my_turtle;


/**
   mc allocation.<br>

   <b>visibility :: <i>public</i></b>

   @param   t_arg                       the type(variable, literal, ...)
   @param   arg                         the value

   @return  the allocated mc
 */
struct s_mc *mc_alloc(mc_t t_arg, gpointer arg)
{
  struct s_mc *mymc = g_new(struct s_mc, 1);

  mymc->t_arg = t_arg;
  mymc->arg = arg;
  mymc->refcnt = 0;

  return(mymc);
}


/**
   mc freeing.<br>

   <b>visibility :: <i>public</i></b>

   @remarks:
            1. we are freeing the mc argument only if it's a literal
	       (we suppose that variables and instructions are referenced 
	        elsewhere)
   @param   mc                          the mc to free
   @return  <none>
 */
void mc_free(struct s_mc *mc)
{
  if(mc != NULL)
    {
      if(--(mc->refcnt) == 0)
	{
	  if(mc->t_arg == MC_LIT)
	    { lit_free(mc->arg, NULL); }
	  else if(mc->t_arg == MC_LIST)
	    { g_slist_foreach(mc->arg, mc_free, NULL); }

	  g_free(mc);
	}
    }
}


/**
   Get the type of an mc argument.<br>

   <b>visibility :: <i>public</i></b>

   @remarks:
            1. we return T_ILLEGAL if the mc as no argument or 
	       T_ILLEGAL/T_NONE typed argument

   @param   mc                          the considered mc
   @return  the type of its argument
 */

/*
 * retourne le type d'une mc (T_ILLEGAL si mc = NULL ou mv->arg = NULL)
 */
inline type_t MC_GET_ARG_TYPE(struct s_mc *mc)
{
  type_t type;
  
  if(mc == NULL || mc->arg == NULL)
    { type = T_ILLEGAL; }
  else
    {
      switch(mc->t_arg)
        {              
          case MC_VAR: type = ((struct s_variable *)mc->arg)->type; break;
          case MC_LIT: type = ((struct s_literal *)mc->arg)->type; break;
          case MC_REF: type = ((struct s_instr *)mc->arg)->type; break;
          default    : type = T_ILLEGAL;
	}
    }

  return(type);
}

inline gpointer MC_GET_ARG_VALUE(struct s_mc *mc)
{
  gpointer value = NULL;
  
  if(mc != NULL && mc->arg != NULL)
    {
      switch(mc->t_arg)
        {              
          case MC_VAR : value = ((struct s_variable *)mc->arg)->value; break;
          case MC_LIT : value = ((struct s_literal *)mc->arg)->value; break;
          case MC_REF : value = ((struct s_instr *)mc->arg)->value; break;
          case MC_LIST: value = (GSList *)mc->arg; break;
	}
    }

  return(value);
}

/**
   Freeing and reinitializing all middle-code informations.<br>

   <b>visibility :: <i>public</i></b>

   @remarks
            1. you must call this function before any code compilation
   @return  <none>
 */
void mc_init(void)
{
  g_slist_foreach(my_turtle.procedures, proc_free, NULL);
  g_slist_free(my_turtle.procedures);
  my_turtle.procedures = NULL;

  /* allocation the main (& default) procedure */
  my_turtle.main_proc = proc_alloc("*:main:*");
  my_turtle.cur_proc = my_turtle.main_proc;
  my_turtle.procedures = g_slist_append(my_turtle.procedures, 
					my_turtle.main_proc);
}


/**
   Entering a new procedure.<br>

   <b>visibility :: <i>public</i></b>

   @remarks
            1. procedures aren't recursives
            1. procedure name must be unique

   @param   name                        the procedure name

   @return  <true> if ok, <false> if a same-name procedure ever exist.
 */
gboolean mc_beginning_procedure(gchar *name)
{
  if(g_slist_find_custom(my_turtle.procedures, name, proc_comparison) != NULL)
    { 
      // error message printing
      // setting error-switch to TRUE
      return(FALSE); 
    }

  my_turtle.cur_proc = proc_alloc(name);
  my_turtle.procedures = g_slist_append(my_turtle.procedures, 
					my_turtle.cur_proc);  
  return(TRUE);
}


/**
   Leaving the current procedure.<br>

   <b>visibility :: <i>public</i></b>

   @remarks:
            1. the default procedure is the main one
   @return  <none>
 */
void mc_ending_procedure(void)
{
  my_turtle.cur_proc = my_turtle.main_proc;
}


/**
   Adding a parameter to the current procedure.<br>

   <b>visibility :: <i>public</i></b>

   @remarks:
            1. the parameters when the procedure will be call shall be in the
               same order
            2. parameter names must be unique

   @param   name                        the parameter name

   @return  <none>
 */
gboolean mc_add_proc_parameter(gchar *name)
{
  struct s_variable *param;

  if(g_slist_find_custom((my_turtle.cur_proc)->parameters, name, 
			 var_comparison) != NULL)
    { 
      // error message printing
      // setting error-switch to TRUE
      return(FALSE); 
    }

  param = var_alloc(name, T_NONE);
  (my_turtle.cur_proc)->parameters = 
    g_slist_append((my_turtle.cur_proc)->parameters, param);
  return(TRUE);
}


/**
   Adding a literal(i.e constant) value.<br>

   <b>visibility :: <i>public</i></b>

   @param   type                        its type
   @param   value                       its value

   @return  the mc-instruction
 */
struct s_mc *mc_add_literal(type_t type, ...)
{
  struct s_literal *lit;
  va_list ap;

  /* on rcupre la valeur du litral */
  va_start(ap, type);
  switch(type)
    {
    case T_INT   : 
      {
	guint value = va_arg(ap, guint);
	lit = lit_alloc(type, value);
	break;
      }
    case T_FLOAT : 
      {
	gdouble value = va_arg(ap, gdouble);
	lit = lit_alloc(type, value);
	break;
      }
    case T_BOOL  : 
      {
	gboolean value = va_arg(ap, gboolean);
	lit = lit_alloc(type, value);
	break;
      }
    case T_STRING: 
      {
	gchar *value = va_arg(ap, gchar *);
	lit = lit_alloc(type, value);
	break;
      }
    default      : lit = lit_alloc(type, NULL);
    }
  va_end(ap);

  return(mc_alloc(MC_LIT, lit));
}


/**
   Adding a variable reference (i.e. used in a right-expression).<br>

   <b>visibility :: <i>public</i></b>

   @remarks:
            1. we must check that the variable exist
	       (local variable, global variable or parameter)
   @param   name                        its name

   @return  the mc-instruction
 */
struct s_mc *mc_add_variable(gchar *varname)
{
  struct s_variable *var;

  var = var_alloc(varname, T_NONE);
  (my_turtle.main_proc)->variables = 
    g_slist_append((my_turtle.main_proc)->variables, var);
  /* !!! we must check that the variable exist, and its type */
  /*var = find_variable(scope, varname, &uvn);
  if(var == NULL)
    { _EMIT_ERROR("variable %s is not declared", varname); } 
  else if(var->type == T_ILLEGAL)
    { 
      _EMIT_ERROR("variable type for %s is unknown", varname);
      var = NULL;
    } 
  */

  return(mc_alloc(MC_VAR, var));
}

/**
 */
struct s_mc *mc_add_variable_reference(gchar *varname)
{
  GSList *varInList;

  /* 1. on recherche les variables comme argument de la fonction */
  varInList = g_slist_find_custom((my_turtle.cur_proc)->parameters, varname, 
				  var_comparison);

  /* 2. on recherche les variables comme variable globale        */
  if(varInList == NULL)
    {
      varInList = g_slist_find_custom((my_turtle.main_proc)->variables, 
				      (gconstpointer) varname, var_comparison);
    }

  if(varInList == NULL)
    { return(NULL); }
  else
    { return(mc_alloc(MC_VAR, varInList->data)); }
}


/**
   Add a expression with no argument (i.e "CS").<br>

   <b>visibility :: <i>public</i></b>

   @param   alloc_mc                    must we allocate a mc
   @param   op                          the operation      
   @return  the resulting mc-expression, or <null> if alloc_mc = FALSE.
 */
struct s_mc *mc_add_noarg_op(gboolean alloc_mc, op_t op)
{
  struct s_instr *instr;
  type_t type;
  
  //itype = type_is_coherent(op);
  type = T_NONE;
  instr = instr_alloc(op, type, NULL, NULL);
  (my_turtle.cur_proc)->instructions = 
    g_slist_append((my_turtle.cur_proc)->instructions, instr);

  return(alloc_mc?mc_alloc(MC_REF, instr):NULL);
}


/**
   Add a unary expression (i.e "TD 10").<br>

   <b>visibility :: <i>public</i></b>

   @param   alloc_mc                    must we allocate a mc
   @param   op                          the operation      
   @param   arg                         the argument (of "struct s_mc *" type)
   @return  the resulting mc-expression, or <null> if alloc_mc = FALSE.
 */
struct s_mc *mc_add_unary_op(gboolean alloc_mc, op_t op, struct s_mc *arg)
{
  struct s_instr *instr;
  type_t type;
  
  //type = type_is_coherent(op, MC_GET_ARG_TYPE(arg));
  type = T_NONE;
  if(type == T_ILLEGAL)
    { 
      //emitting error
      /*_EMIT_ERROR("invalid opration %s on type %s", 
		  op_to_string(op), type_to_string(MC_GET_ARG_TYPE(arg)));
      */
    }

  instr = instr_alloc(op, type, arg, NULL);
  (my_turtle.cur_proc)->instructions = 
    g_slist_append((my_turtle.cur_proc)->instructions, instr);

  return(alloc_mc?mc_alloc(MC_REF, instr):NULL);
}


/**
   Add a binary expression (i.e "a + b").<br>

   <b>visibility :: <i>public</i></b>

   @param   alloc_mc                    must we allocate a mc
   @param   op                          the operation      
   @param   arg1                        the first argument
   @param   arg2                        the second argument
   @return  the resulting mc-expression, or <null> if alloc_mc = FALSE.
 */
struct s_mc *mc_add_binary_op(gboolean alloc_mc, op_t op, 
			      struct s_mc *arg1, struct s_mc *arg2)
{
  struct s_instr *instr;
  type_t type;
  
  //type = type_is_coherent(op, MC_GET_ARG_TYPE(arg1), MC_GET_ARG_TYPE(arg2));
  type = T_NONE;
  if(type == T_ILLEGAL)
    { 
      //emitting error
      /*_EMIT_ERROR("invalid opration %s on types (%s, %s)", 
		  op_to_string(op), 
		  type_to_string(MC_GET_ARG_TYPE(arg1)), 
		  type_to_string(MC_GET_ARG_TYPE(arg2)));
      */
    }

  instr = instr_alloc(op, type, arg1, arg2);
  (my_turtle.cur_proc)->instructions = 
    g_slist_append((my_turtle.cur_proc)->instructions, instr);

  return(alloc_mc?mc_alloc(MC_REF, instr):NULL);
}


/**
   Update the destination of a branch instruction.<br>

   <b>visibility :: <i>public</i></b>

   @remarks:
            1. the destination of a branch is the first argument of the
               branch instruction
	       (the second is the optional condition)

   @param   instr                       the instruction to update
   @param   dest                        the destination

   @return  <none>
 */
void mc_update_branch(struct s_instr *instr, struct s_instr *dest)
{
  //gint index;
  //struct s_literal *lit;

  //index = g_slist_index((my_turtle.cur_proc)->instructions, dest);
  //lit = lit_alloc(T_INT, (gpointer) index);
  //instr->arg1 = mc_alloc(MC_LIT, lit);

  instr->arg1 = mc_alloc(MC_REF, dest);
}


gpointer mc_find_procedure(gchar *name)
{
  GSList *elt;

  elt = g_slist_find_custom(my_turtle.procedures, name, proc_comparison);
  return(elt == NULL?NULL:elt->data);
}

#ifdef DEBUG1
void mc_debug(void)
{
  gint pos;

  //  gtk_notebook_set_page(GTK_NOTEBOOK(my_turtle.notebook), 1);
  debug_clear();
  debug_append("*** Starting debug\n");
  g_slist_foreach(my_turtle.procedures, proc_debug, NULL);
  
  debug_append("*** Ending debug\n");
}
#endif
/*

   <b>visibility :: <i>public</i></b>

   @param   instr                       the instruction to update
   @param   argno                       the number of the argument (0, 1 or 2)
   @param   niouval                     the new argument value

   @return  the old value, or <null> if error
 */
/*struct s_mc *mc_update_arg(guint index, gushort argno, struct s_mc *niouval)
{
  struct s_mc *oldval = NULL;
  struct s_instr *instr;

  instr = g_slist_nth_data((my_turtle.cur_proc)->instructions , index);
*/  /* not found instruction (invalid index) *//*
  if(instr == NULL)
    { return(NULL); }

	       */  /* updating the argument *//*
  switch(argno)
    {
    case 0: 
      oldval = instr->arg1;
      oldtval = instr->targ1;
      instr->arg1 = niouval;
      instr->targ1 = nioutval;
      break;
    case 1: 
      oldval = instr->arg2;
      oldtval = instr->targ2;
      instr->arg2 = niouval;
      instr->targ2 = nioutval;
      break;
    case 2: 
      oldval = instr->arg3;
      oldtval = instr->targ3;
      instr->arg3 = niouval;
      instr->targ3 = nioutval;
    }

  return(oldval);
}
*/
